#pragma once
/** Sysvars library
 * Copyright (c) 2018-2023 Pavel Brychta, pablo@xpablo.cz
 *
 * Key-value store/load library with JSON file using internal filesystem
 **/

#include "definefs.hpp"
#include <Arduino.h>
#include <ArduinoJson.h>
#include <FS.h>

enum {
    SV_OK = 0,
    SV_UNSPECIFIED_ERROR = -1,
    SV_FILE_NOT_FOUND = -2,
    SV_DESERIALIZE_ERROR = -3,
    SV_WRONG_DATA_TYPE = -4,
    SV_FILE_WRITE_ERROR = -5,
    SV_KEY_NOT_FOUND = -6,
};

#if not defined(SYSVAR_FILE)
    #define SYSVAR_FILE "/cfg.json"
#endif

#if not defined(SYSVAR_DOCUMENT_SIZE)
    #define SYSVAR_DOCUMENT_SIZE (4096)
#endif

/**
 * @brief      Get value
 *
 * @param[in]  name    Name of the key
 * @param      result  Status variable (SV_...)
 *
 * @tparam     T       result type
 *
 * @return     Value for the key given
 */
template <typename T>
T svGetV(const String & name, int * result = nullptr)
{
    T rv{};
    int res;
    DynamicJsonDocument doc(SYSVAR_DOCUMENT_SIZE);
    File f;

    f = USEDFS.open(F(SYSVAR_FILE), "r");
    if (f) {
        DeserializationError err = deserializeJson(doc, f);
        if (err) {
            res = SV_DESERIALIZE_ERROR; // wrong JSON file (malformed)
        } else {
            if (!doc.containsKey(name)) {
                res = SV_KEY_NOT_FOUND; // key not found
            } else {
                rv = doc[name].as<T>();
                res = SV_OK;
            }
        }
    } else {
        res = SV_FILE_NOT_FOUND; // file not found
    }
    if (result)
        *result = res; // result code if we are interested
    return rv;
}

/**
 * @brief      Get value
 *
 * @param[in]  name    Name of the key
 * @param      result  Status variable (SV_....)
 * @param[in]  fname   Used JSON file name
 *
 * @tparam     T       Result type
 *
 * @return     Value for the key given
 */
template <typename T>
T svGetV(const String & name, int * result, const String & fname)
{
    T rv{};
    int res;
    DynamicJsonDocument doc(SYSVAR_DOCUMENT_SIZE);
    File f;

    f = USEDFS.open(fname, "r");
    if (f) {
        DeserializationError err = deserializeJson(doc, f);
        if (err) {
            res = SV_DESERIALIZE_ERROR; // wrong (malformed) JSON file
        } else {
            if (!doc.containsKey(name)) {
                res = SV_KEY_NOT_FOUND; // key not found
            } else {
                rv = doc[name].as<T>();
                res = SV_OK;
            }
        }
    } else {
        res = SV_FILE_NOT_FOUND; // file not found
    }
    if (result)
        *result = res; // result code if we are interested
    return rv;
}

/**
 * @brief      Store value for the key
 *
 * @param[in]  name   Key name
 * @param[in]  value  New value
 *
 * @tparam     T      Type of the value
 *
 * @return     Result (SV_...)
 */
template <typename T>
int svSetV(const String & name, T value)
{
    int res;
    DynamicJsonDocument doc(SYSVAR_DOCUMENT_SIZE);
    File f;

    f = USEDFS.open(F(SYSVAR_FILE), "r");
    if (f) {
        DeserializationError err = deserializeJson(doc, f);
        if (err) {
            res = SV_DESERIALIZE_ERROR; // wrong (malformed) JSON file
        }
        f.close();
    }
    doc[name] = value;

    f = USEDFS.open(SYSVAR_FILE, "w");
    if (f) {
        serializeJson(doc, f);
        res = SV_OK;
    } else {
        res = SV_FILE_WRITE_ERROR;
    }
    return res;
}

/**
 * @brief      Store value for the key
 *
 * @param[in]  name   Key name
 * @param[in]  value  New value
 * @param[in]  fname  JSON file name
 *
 * @tparam     T      Type of the value
 *
 * @return     Result (SV_...)
 */
template <typename T>
int svSetV(const String & name, T value, const String & fname)
{
    int res = SV_UNSPECIFIED_ERROR;
    DynamicJsonDocument doc(SYSVAR_DOCUMENT_SIZE);
    File f;

    f = USEDFS.open(fname, "r");
    if (f) {
        DeserializationError err = deserializeJson(doc, f);
        if (err) {
            res = SV_DESERIALIZE_ERROR; // wrong (malformed) JSON file
        }
        f.close();
    }
    doc[name] = value;

    f = USEDFS.open(fname, "w");
    if (f) {
        serializeJson(doc, f);
        res = SV_OK;
    } else {
        res = SV_FILE_WRITE_ERROR;
    }

    return res;
}
